#ifndef __MATRIX4x4_H
#define __MATRIX4x4_H

#include <iostream>
//

namespace GtpVisibilityPreprocessor {

class Vector3;



class Matrix4x4
{
public:
  float x[4][4]; // first index is column [x], the second is row [y]

  Matrix4x4() { }

  // here xXY - 'X' is row, and 'Y' is column - classical mathematical notation
  Matrix4x4(float x11, float x12, float x13, float x14,
	   float x21, float x22, float x23, float x24,
	   float x31, float x32, float x33, float x34,
	   float x41, float x42, float x43, float x44);

  Matrix4x4(const Vector3 &a, const Vector3 &b, const Vector3 &c);


  // Assignment operators
  Matrix4x4& operator+= (const Matrix4x4 &A); 	// add-to
  Matrix4x4& operator-= (const Matrix4x4 &A); 	// subtract-from
  Matrix4x4& operator*= (const Matrix4x4 &A); 	// multiply by matrix
  Matrix4x4& operator*= (float A); 		// scale by scalar

  // Fundamental operations
  /// Invert the matrix .. returns 0 = regular
  int Invert();		
  /// Transpose the matrix
  void Transpose();	

  void
  SetColumns(const Vector3 &a, const Vector3 &b, const Vector3 &c);

  float Det3x3() const {
    return (x[0][0]*x[1][1]*x[2][2] + \
	    x[1][0]*x[2][1]*x[0][2] + \
	    x[2][0]*x[0][1]*x[1][2] - \
	    x[2][0]*x[1][1]*x[0][2] - \
	    x[0][0]*x[2][1]*x[1][2] - \
	    x[1][0]*x[0][1]*x[2][2]);
  }
  

  friend Matrix4x4 Invert(const Matrix4x4 &M);	// Invert a given matrix
  friend Matrix4x4 Transpose(const Matrix4x4 &M);	// Transpose a given matrix

  // Create various types of matrix.
  friend Matrix4x4 IdentityMatrix();
  friend Matrix4x4 ZeroMatrix();
  friend Matrix4x4 TranslationMatrix(const Vector3 &Location);
  friend Matrix4x4 RotationXMatrix(float Angle);
  friend Matrix4x4 RotationYMatrix(float Angle);
  friend Matrix4x4 RotationZMatrix(float Angle);
  friend Matrix4x4 RotationYPRMatrix(float Yaw, float Pitch, float Roll);
  // about axis 'axis' by angle 'Angle'
  friend Matrix4x4 RotationAxisMatrix(const Vector3 &axis, float Angle);
  // create the rotation matrix that rotates 'vecFrom' to 'vecTo'
  friend Matrix4x4 RotationVectorsMatrix(const Vector3 &vecFrom,
					 const Vector3 &vecTo);
  
  friend Matrix4x4 ScaleMatrix(float X, float Y, float Z);
  friend Matrix4x4 GenRotation(const Vector3 &x, const Vector3 &y,
			     const Vector3 &z);
  friend Matrix4x4 QuadricMatrix(float a, float b, float c, float d, float e,
			      float f, float g, float h, float j, float k);
  // returns matrix for transforming normal
  friend Matrix4x4 NormalTransformMatrix(const Matrix4x4 &M);

  friend Matrix4x4 MirrorX();
  friend Matrix4x4 MirrorY();
  friend Matrix4x4 MirrorZ();
  friend Matrix4x4 RotationOnly(const Matrix4x4 &x);

  // Binary operators
  friend Matrix4x4 operator+ (const Matrix4x4 &A, const Matrix4x4 &B);
  friend Matrix4x4 operator- (const Matrix4x4 &A, const Matrix4x4 &B);
  friend Matrix4x4 operator* (const Matrix4x4 &A, float B);
  friend Matrix4x4 operator* (const Matrix4x4 &A, const Matrix4x4 &B);

  // friends returning Vector3
  friend Vector3 operator*(const Matrix4x4 &M, const Vector3 &v);
  friend Vector3 RotateOnly(const Matrix4x4 &M, const Vector3 &v);
  friend Vector3 TransformNormal(const Matrix4x4 &M, const Vector3 &v);
  friend Vector3 GetTranslation(const Matrix4x4 &M);

  // Construct rotation description according VRML'97 specification
  //  const CVector4D SFRotation(void) const;

  

  // Overloaded output operator.
  friend std::ostream& operator<< (std::ostream &s, const Matrix4x4 &M);
};

}

#endif
